TITLE = "Title"
TITLE_EXPLAIN = "Give the issue an explicit title."
DESCRIPTION = "Description"
DESCRIPTION_EXPLAIN = "Detail the issue you have observed."
FEEDBACK_SAVE_BUTTON = "Save feedback"
LOCAL_SAVE = "Feedback saved."
LOCAL_SAVE_HTML = f'<p style="color:Green;">{LOCAL_SAVE}</p>'
PLATFORM_SAVE = "Feedback saved to Trubrics."
PLATFORM_SAVE_HTML = f'<p style="color:Green;">{PLATFORM_SAVE}</p>'
FEEDBACK_NOT_SAVED = "Please specify both a feedback title and a description."
FEEDBACK_NOT_SAVED_HTML = '<p style="color:Red;">Please specify a feedback title and a description.</p>'
USER_EMAIL = "User email"
USER_PASSWORD = "User password"
